## @file
# This file is used to save global datas used by ECC tool
#
# Copyright (c) 2008 - 2018, Intel Corporation. All rights reserved.<BR>
# SPDX-License-Identifier: BSD-2-Clause-Patent
#

##
# Import Modules
#
import edk2basetools.Common.LongFilePathOs as os

gWorkspace = ''
gTarget = ''
gConfig = None
gDb = None
gIdentifierTableList = []
gCFileList = []
gHFileList = []
gUFileList = []
gException = None
